/*
 * crash-manager
 * Copyright (c) 2019 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#include <assert.h>
#include <dirent.h>
#include <fcntl.h>
#include <iniparser.h>
#include <stdbool.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "config.h"
#include "defs.h"
#include "log.h"
#include "util.h"

static const char *const report_type_strmap[] = {
	[REP_TYPE_INFO] = "INFO",
	[REP_TYPE_FULL] = "FULL",
};

static const char *report_type_to_str(enum ReportType report_type)
{
	if (report_type < 0 || report_type >= ARRAY_SIZE(report_type_strmap))
		return NULL;
	return report_type_strmap[report_type];
}

enum ReportType report_type_from_str(const char *report_type_str)
{
	assert(report_type_str);

	for (int i = 0; i < (int)ARRAY_SIZE(report_type_strmap); i++) {
		if (0 == strcmp(report_type_str, report_type_strmap[i]))
			return (enum ReportType)i;
	}

	return REP_TYPE_INVALID;
}

static bool config_load_from_dict(config_t *c, dictionary *ini)
{
	assert(c);

	char *str = iniparser_getstring(ini, CRASH_SECTION ":CrashRootPath", NULL);
	if (str) {
		char *crash_root_path = strdup(str);
		if (!crash_root_path) {
			_E("config: Unable to set CrashRootPath - aborting");
			return false;
		}

		free(c->crash_root_path);
		c->crash_root_path = crash_root_path;
	}

	/* strdup() can technically fail, but we don't mind. It is better to
	 * create a report without the extra script than to abort completely. */
	str = iniparser_getstring(ini, CRASH_SECTION ":ExtraScript", NULL);
	if (str) {
		char *extra_script = strdup(str);
		if (extra_script) {
			free(c->extra_script);
			c->extra_script = extra_script;
		}
	}

	str = iniparser_getstring(ini, CRASH_SECTION ":ReportType", NULL);
	if (str) {
		int type = report_type_from_str(str);
		if (report_type_to_str(type))
			c->report_type = type;
	}

#define UPDATE(where, type, key) where = iniparser_get##type(ini, CRASH_SECTION ":" key, where)

	UPDATE(c->system_max_use, int, "SystemMaxUse");
	UPDATE(c->max_retention_sec, int, "MaxRetentionSec");
	UPDATE(c->max_crash_dump, int, "MaxCrashDump");
	UPDATE(c->dump_core, boolean, "DumpCore");
	UPDATE(c->allow_zip, boolean, "AllowZip");

#undef UPDATE

	return true;
}

static bool config_load_from_path(config_t *c, const char *const path)
{
	assert(path);

	dictionary *ini = iniparser_load(path);
	if (!ini) {
		_E("Failed to load config file %s", path);
		return false;
	}

	bool ret = config_load_from_dict(c, ini);
	iniparser_freedict(ini);

	return ret;
}

static int entry_filter(const struct dirent *e)
{
	assert(e);

	const char *const conf_suffix = ".conf";
	const char *const name = e->d_name;
	int len = strlen(name);

	if (e->d_type != DT_REG || len <= strlen(conf_suffix))
		return 0;

	// accept only files ending with predefined suffix
	return strcmp(name + len - strlen(conf_suffix), conf_suffix) == 0;
}

static bool config_load_from_dir_prefix(config_t *c, const char *const dir_prefix)
{
	assert(dir_prefix);

	char dir_path[PATH_MAX];
	int ret = snprintf(dir_path, sizeof(dir_path), "%s.d", dir_prefix);
	if (ret < 0 || ret >= PATH_MAX) {
		_W("config: internal error while trying to prepare config dir path");
		return false;
	}

	int fd = open(dir_path, O_RDONLY | O_DIRECTORY);
	if (fd < 0 && errno == ENOENT)
		return true;
	if (fd < 0) {
		_E("config: Unable to access config directory at %s: %m", dir_path);
		return false;
	}

	struct dirent **entries = NULL;
	int n = scandirat(fd, ".", &entries, entry_filter, alphasort);
	if (n < 0)
		goto out;

	for (int i = 0; i < n; ++i) {
		char file_path[PATH_MAX];
		const char *fname = entries[i]->d_name;
		ret = snprintf(file_path, sizeof(file_path), "%s/%s", dir_path, fname);
		if (ret < 0 || ret >= PATH_MAX) {
			_W("config: internal error while trying to prepare for reading %s config file", fname);
			continue;
		}
		_D("config: reading additional configuration file from %s", file_path);
		(void)config_load_from_path(c, file_path);
	}

out:
	free(entries);
	close(fd);
	return true;
}

static bool config_apply_defaults(config_t *c)
{
	assert(c);

	memset(c, 0, sizeof(*c));

	c->crash_root_path = strdup(CRASH_ROOT_PATH);
	c->report_type = REP_TYPE_FULL;
	c->system_max_use = SYSTEM_MAX_USE;
	c->system_keep_free = SYSTEM_KEEP_FREE;
	c->max_retention_sec = MAX_RETENTION_SEC;
	c->max_crash_dump = MAX_CRASH_DUMP;
	c->dump_core = DUMP_CORE;
	c->allow_zip = ALLOW_ZIP;

	return c->crash_root_path != NULL;
}

static void config_dump(config_t *c)
{
	assert(c);

	_D("config: crash_root_path = %s\n"
	   "config: extra_script = %s\n"
	   "config: report_type = %s\n"
	   "config: system_max_use = %d\n"
	   "config: system_keep_free = %d\n"
	   "config: max_retention_sec = %d\n"
	   "config: max_crash_dump = %d\n"
	   "config: dump_core = %d\n"
	   "config: allow_zip = %d\n",
	   c->crash_root_path,
	   c->extra_script,
	   report_type_to_str(c->report_type),
	   c->system_max_use,
	   c->system_keep_free,
	   c->max_retention_sec,
	   c->max_crash_dump,
	   c->dump_core,
	   c->allow_zip);
}

bool config_init(config_t *c, const char *const path)
{
	if (!config_apply_defaults(c) || !config_load_from_path(c, path)) {
		_E("config: Unable to initialize configuration");
		return false;
	}

	(void)config_load_from_dir_prefix(c, path);

	config_dump(c);

	return true;
}

void config_free(config_t *c)
{
	assert(c);

	free(c->crash_root_path);
	free(c->extra_script);

	memset(c, 0, sizeof(*c));
}
